jpc <- function(dat, j = 3, k = 1, diss=FALSE, verbose=FALSE) {

    pair.dist <- function(dat) {
        dist <- matrix(0,nr=nrow(dat), nc=nrow(dat))
        dat <- as.matrix(dat)
        retlist <- .C("pairdist", PACKAGE="jpc",
        as.double(dat),
        as.integer(nrow(dat)), as.integer(ncol(dat)),
        as.double(dist))
        dist <- as.data.frame(matrix(retlist[[4]], nr=nrow(dat)))
        names(dist) <- 1:nrow(dat)
        dist
    }
    in.clus <- function(cluslist,obj) {
        # returns the number of the cluster the
        # specified object was found in. If not
        # in any of the clusters returns 0
        cluslist[ cluslist$objid == obj, ]$clusid
    }
    add.clus <- function(cluslist, obj, whichclus) {
        cluslist[ cluslist$objid == obj, ]$clusid <- whichclus
        cluslist
    }
    new.clus <- function(cluslist, objs) {
        newid <- max(unique(cluslist$clusid)) + 1
        idxs <- which(cluslist$objid %in% objs)
        cluslist[ idxs, ]$clusid <- newid
        cluslist
    }
    len.clus <- function(cluslist, clusid) {
        nrow(cluslist[ cluslist$clusid == clusid, ])
    }
    num.clus <- function(cluslist) {
        # since initially objects with clusid == 0 indicate no membeship
        # in any cluster, we should reduce th ecount ofunique clusid's by 1
        # if 0 is present at all
        u <- unique(cluslist$clusid)
        if (any(u == 0)) { length(u)-1 }
        else { length(u)}

    }
    merge.clus <- function(cluslist,cid1, cid2) {
        # we merge the cluster with higher id into the one with lower id
        # parsimony :)
        mincid <- min(cid1,cid2)
        maxcid <- max(cid1,cid2)
        cluslist[ cluslist$clusid == maxcid, ]$clusid <- mincid
        cluslist
    }

    ##################################################################################

    
    k <- as.integer(k)
    j <- as.integer(j)
    if (j == 0 || k == 0) {
        stop("j and k > 0")
    }

    if (diss) {
        if (ncol(dat) != nrow(dat)) {
            stop("If a distance matrix is supplied it should be square symmetric")
        }
        dist <- dat
    } else {
        dist <- pair.dist(dat)
    }

    if (verbose) {print('Starting jNN')}

    jnn <- list()
    for (i in 1:nrow(dist)) {
        tmp <- as.data.frame(cbind(as.numeric(names(dist)), dist[,i]))
        tmp <- tmp[sort.list(tmp$V2),]
        jnn[[i]] <- tmp[2:(2+j-1),1] # since the first element will always be 0
    }
    jnn <- lapply(jnn,sort)

#    dat <- matrix(0,nr=10,nc=4)
#    jnn <- list()
#    jnn[[1]] <- c(2,3,4)
#    jnn[[2]] <- c(1,3,4)
#    jnn[[3]] <- c(1,2,4)
#    jnn[[4]] <- c(5,6,7)
#    jnn[[5]] <- c(4,6,7)
#    jnn[[6]] <- c(4,5,7)
#    jnn[[7]] <- c(4,5,6)
#    jnn[[8]] <- c(4,5,9)
#    jnn[[9]] <- c(2,8,5)
#    jnn[[10]] <- c(1,4,9)

    if (verbose) {print('starting Clustering')}

    # make the cluster data structure
    cluslist <- data.frame( 1:nrow(dat), rep(0,nrow(dat)) )
    names(cluslist) <- c('objid','clusid')
    cluslist <- new.clus(cluslist, 8)

    # look at all pairs. 
    for (i in 1:length(jnn)) {
        for (j in 1:length(jnn)) {

            if (i == j) {next}

            # do the tests. Both tests have to be passed
            # for i & j to be in the same cluster
            flag1 = FALSE
            flag2 = FALSE

            if (i %in% jnn[[j]] && j %in% jnn[[i]]) {
                flag1 = TRUE
            } 
            ncommon <- sum( jnn[[i]] %in% jnn[[j]] )
            if (ncommon >= k) {
                flag2 = TRUE
            }


            # cluster management!
            c1 <- in.clus(cluslist,i)
            c2 <- in.clus(cluslist,j)
            if (flag1 && flag2) { # Passed the tests
                #print(paste('(PASS) membership: i,j',i,j,'cids',c1,c2))

                # if i and j are alredy in the same cluster carry on
                if (c1 == c2 && c1 > 0 && c2 > 0) { 
                    #print(paste(i,j,' already in same cluster. Next'))
                    next
                }

                # if neither are in a cluster then we should place them
                # together in a new cluster and carry on
                if (c1 == 0 && c2 == 0) {
                    #print(paste('  ',i,j,' not in any cluster. Add them to a new one'))
                    cluslist <- new.clus(cluslist, c(i,j))
                    next
                }

                # if i or j is in a cluster and j or i is not
                # then add the molecule with no assignment to the cluster
                # of the other molecule
                if (c1 != 0 && c2 == 0) { 
                    #print(paste('  add ',j,' to ',i,'\'s cluster (',c1,')'))
                    cluslist = add.clus(cluslist, j, c1)
                    next
                }
                if (c1 == 0 && c2 != 0) { 
                    #print(paste('  add ',i,' to ',j,'\'s cluster (',c2,')'))
                    cluslist = add.clus(cluslist, i, c2)
                    next
                }

                # if i & j are in seperate clusters we then need to check for
                # singletons.
                if (c1 != c2 && c1 > 0 && c2 > 0) {

                    # at this point I'm not sure. If i and j are in two seperate clusters
                    # and are singletons then we simply merge the two clusters. 
                    # However if they are in two seperate clusters and are *not* singeltons we face
                    # a problem - since if i and j are in this section they have passed the 
                    # two tests and should be in the same group. So why do we find them in two 
                    # seperate clusters?
                    #
                    # In any case given that i and j are in two seperate non singleton clusters
                    # we can either skip them and let them remain where they are. Alternatively
                    # we can merge the two clusters. This is based on the fact that 
                    #           (L,M) && (L,N) ==> (M,N)
                    # (where (L,M) imply L & M are in the same cluster)
                    # so if we see i is in a cluster with a & b and j is in a cluster with x & y
                    # and (i,j) then since (a,j) and (b,j) ==> (a,i) and (b,i) and so on.
                    # This seems to make sense, so effectively, if i and j are in two seperate clustes
                    # we just merge them without paying attention to singeltons
                    cluslist <- merge.clus(cluslist,c1,c2)
                }
            }    
            else { 
                #print(paste('(FAIL) membership: i,j',i,j,'cids',c1,c2))
                if (c1 == 0) {
                    #print(paste(' add ',i,'to new cluster'))
                    cluslist = new.clus(cluslist, c(i))
                }
                if (c2 == 0) {
                    #print(paste(' add ',j,'to new cluster'))
                    cluslist = new.clus(cluslist, c(j))
                }
            }
        }
    }

    # renumber clusters to start from 1 and increase sequentially
    u <- sort(unique(cluslist$clusid))
    nu <- 1:length(u)
    for (i in 1:length(u)) {
        cluslist[ cluslist$clusid == u[i], ]$clusid <- nu[i]
    }
    list(clus=cluslist,nclus=length(nu))
}




